from .pkg_data.crypto import base64_to_a32, base64_url_decode, decrypt_attr, decrypt_key, a32_to_str, AES, get_chunks, str_to_a32
from tenacity import retry, wait_exponential, retry_if_exception_type
from .pkg_data.mega import Counter, logger
from .pkg_data.errors import RequestError
from omnitools import def_template
from .pkg_data import Mega as _Mega
from pathlib import Path
import threadwrapper
import threading
import traceback
import tempfile
import requests
import random
import shutil
import time
import copy
import json
import re
import os
# from typing import *


class Mega(_Mega):
    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    #     self.domain = "mega.nz"

    def download(self, file, dest_path=None, dest_filename=None):
        raise NotImplementedError

    @staticmethod
    def parse_folder_url(url: str) -> tuple:
        folder_url_format = [
            re.compile(r"mega.[^/]+/folder/([0-z-_]+)#([0-z-_]+)(?:/folder/([0-z-_]+))*"),
            re.compile(r"mega.[^/]+/#F!([0-z-_]+)[!#]([0-z-_]+)(?:/folder/([0-z-_]+))*")
        ]
        m = re.search(folder_url_format[0], url)
        if not m:
            m = re.search(folder_url_format[1], url)
        if not m:
            print("Not a valid URL")
            return tuple()
        root_folder = m.group(1)
        key = m.group(2)
        # You may want to use m.group(-1)
        # to get the id of the subfolder
        return (root_folder, key)
        
    @staticmethod
    def decrypt_node_key(key_str: str, shared_key: str) -> tuple:
        encrypted_key = base64_to_a32(key_str.split(":")[1])
        return decrypt_key(encrypted_key, shared_key)

    @retry(retry=retry_if_exception_type(RuntimeError),
           wait=wait_exponential(multiplier=2, min=2, max=60))
    def get_nodes_in_folder(self, root_folder: str) -> dict:
        data = [{"a": "f", "c": 1, "ca": 1, "r": 1}]
        response = requests.post(
            "{}://g.api.{}/cs".format(
                self.schema,
                self.domain
            ),
            params={
                "id": self.sequence_num,
                "n": root_folder
            },
            data=json.dumps(data)
        )
        self.sequence_num += 1
        json_resp = response.json()
        if isinstance(json_resp, int):
            print("no files in folder, probably expired")
            return []
        return json_resp[0]["f"]

    @retry(retry=retry_if_exception_type(RuntimeError),
           wait=wait_exponential(multiplier=2, min=2, max=60))
    def get_file_data_in_folder(self, file_id, root_folder, proxies=None):
        data = [{"a": "g", "g": 1, "n": file_id}]
        response = requests.post(
            "{}://g.api.{}/cs".format(
                self.schema,
                self.domain
            ),
            params={
                "id": self.sequence_num,
                "n": root_folder
            },
            data=json.dumps(data),
            timeout=self.timeout,
            proxies=proxies
        )
        self.sequence_num += 1
        return response.json()[0]

    @retry(retry=retry_if_exception_type(RuntimeError),
           wait=wait_exponential(multiplier=2, min=2, max=60))
    def get_file_data(self, file_id, proxies=None):
        response = requests.post(
            '{}://g.api.{}/cs'.format(self.schema, self.domain),
            params={'id': self.sequence_num},
            data=json.dumps([{
                'a': 'g',
                'g': 1,
                'p': file_id
            }]),
            timeout=self.timeout,
            proxies=proxies
        )
        self.sequence_num += 1
        return response.json()[0]

    def get_files_in_folder(self, url: str, verbose: bool = False):
        folder_paths = {}
        (root_folder, shared_enc_key) = self.parse_folder_url(url)
        shared_key = base64_to_a32(shared_enc_key)
        nodes = self.get_nodes_in_folder(root_folder)
        items = []
        for i, node in enumerate(nodes):
            key = self.decrypt_node_key(node["k"], shared_key)
            if node["t"] == 0:
                k = (key[0] ^ key[4], key[1] ^ key[5], key[2] ^ key[6], key[3] ^ key[7])
            elif node["t"] == 1:
                k = key
            # noinspection PyUnboundLocalVariable
            attrs = decrypt_attr(base64_url_decode(node["a"]), k)
            file_name = attrs["n"]
            file_id = node["h"]
            if verbose:
                print("\r", i+1, len(nodes), "File" if node["t"] == 0 else "Folder", file_name, flush=True)
            if node["t"] == 0:
                json_resp = self.get_file_data_in_folder(file_id, root_folder)
                json_resp["path"] = folder_paths[node["p"]]
                items.append([file_id, file_name, node, key, json_resp])
            elif node["t"] == 1:
                parent_node = node["p"]
                if parent_node not in folder_paths:
                    folder_paths[parent_node] = "/"
                folder_paths[file_id] = folder_paths[parent_node]+file_name+"/"
                # path = "/" + file_name
                # if parent_node in folder_paths:
                #     path = folder_paths[parent_node]+"/"+file_name
                # folder_paths[file_id] = path
        return items
    
    def download_files_in_folder(self, url: str, dest_path: str = None, serial_download=True, parallel_download=False, parallel_threads=2**4, verbose: bool = False, proxies=None, stop_on_exception=False) -> list:
        proxies = proxies or [None]
        items = self.get_files_in_folder(url)
        if not dest_path:
            dest_path = ""
        else:
            dest_path += "/"
        output_paths = []
        for i, item in enumerate(items):
            output_paths.append(self.download_file_in_folder(url, items[i], len(items), dest_path, i, serial_download, parallel_download, parallel_threads, verbose, proxies, stop_on_exception))
        return output_paths

    def download_file_in_folder(self, url: str, item: list, l: int, dest_path: str = None, sequence: int = 0, serial_download=True, parallel_download=False, parallel_threads=2**4, verbose: bool = False, proxies=None, stop_on_exception=False) -> str:
        proxies = proxies or [None]
        file_key = tuple(item[3])
        file_data = item[4]
        k = (
            file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
            file_key[2] ^ file_key[6], file_key[3] ^ file_key[7]
        )
        iv = file_key[4:6] + (0, 0)
        meta_mac = file_key[6:8]
        if parallel_download:
            return self.parallel_download(item[0], url, dest_path, "", file_data, k, iv, meta_mac, parallel_threads, verbose, proxies, stop_on_exception)
        elif serial_download:
            if verbose:
                print("\r", sequence+1, l, file_data["path"][1:] if "path" in file_data else "", item[1], file_data["s"], flush=True)
            return self.download_file(dest_path, "", file_data, k, iv, meta_mac, verbose, proxies)

    def parallel_download(self, file_id, root_link, dest_path: str, dest_filename: str, file_data: dict, k: tuple, iv: tuple, meta_mac: tuple, parallel_threads=2**4, verbose=True, proxies=None, stop_on_exception=False):
        proxies = proxies or [None]
        proxies = copy.deepcopy(proxies)
        random.SystemRandom().shuffle(proxies)
        if verbose:
            print(file_id, root_link, file_data)
        file_path = file_data["path"][1:] if "path" in file_data else ""
        is_dir = "/folder/" in root_link or "/#F!" in root_link
        root_folder = None
        if is_dir:
            root_folder = self.parse_folder_url(root_link)[0]
            file_data = self.get_file_data_in_folder(file_id, root_folder)
            if verbose:
                print("folder")
        else:
            file_data = self.get_file_data(file_id)
            if verbose:
                print("file")
        attribs = base64_url_decode(file_data["at"])
        attribs = decrypt_attr(attribs, k)
        if not dest_path:
            dest_path = ""
        else:
            dest_path += "/"
        if dest_filename:
            file_name = dest_filename
        else:
            file_name = attribs["n"]
        file_size = file_data["s"]
        if verbose:
            print(file_path, file_name, file_size)
        parallel_threads = min(parallel_threads, len(proxies))
        if verbose:
            print("using threads:", parallel_threads)
        tws = [threadwrapper.ThreadWrapper(threading.Semaphore(2**0)) for _ in range(0, parallel_threads)]
        cache = {}
        retry = {}
        result = {}
        timeout_blacklist = []
        chunks = list(get_chunks(file_size))
        l = len(chunks)
        done = False
        index = 0
        exploded = False
        downloaded = 0
        completed = 0
        temp_file_size = 0
        temp_pieces_fp = os.path.join(tempfile.gettempdir(), "megapy", "id", file_id)
        os.makedirs(temp_pieces_fp, exist_ok=True)
        default_tempfile_dir = os.path.join(temp_pieces_fp, "..", "..")
        default_tempfile_file_dir = os.path.join(default_tempfile_dir, "file")
        resume_fp = os.path.join(temp_pieces_fp, "..", "..", "resume.txt")
        os.makedirs(default_tempfile_file_dir, exist_ok=True)
        if file_size < min(4*1024*1024*1024, shutil.disk_usage(default_tempfile_file_dir).free-0.1*1024*1024*1024):
            default_tempfile_dir = default_tempfile_file_dir
        if os.path.isfile(resume_fp):
            resume = open(resume_fp, "rb").read().decode()
            if os.path.isfile(resume):
                temp_file_size = os.path.getsize(resume)
                downloaded = temp_file_size
                completed = temp_file_size
                temp_file_size = temp_file_size
        __file_size = 0
        l_c = 0
        _i = 0
        prev_downloaded = 0
        prev_completed = 0
        download_rate_mb = 0
        complete_rate_mb = 0
        prev_time = time.time()
        def progress():
            nonlocal download_rate_mb
            nonlocal complete_rate_mb
            nonlocal prev_time
            nonlocal prev_downloaded
            nonlocal prev_completed
            ctime = time.time()
            diff = ctime-prev_time
            if diff > 2:
                download_rate_mb = (downloaded-prev_downloaded)/(diff or 1)/1024/1024
                complete_rate_mb = (completed-prev_completed)/(diff or 1)/1024/1024
                prev_time = ctime
                prev_downloaded = downloaded
                prev_completed = completed
            print("\r", "handler   downloading piece {}/{} [{:.2f}%]   downloaded {:.2f}% [{:.2f} MB/s]   completed {:.2f}% [{:.2f} MB/s]".format(
                _i+1,
                l,
                (_i+1)/l*100,
                downloaded/(file_size or 1)*100,
                download_rate_mb,
                completed/(file_size or 1)*100,
                complete_rate_mb,
                ), end="", flush=True)
        def file_handler():
            nonlocal index
            nonlocal done
            nonlocal exploded
            nonlocal completed
            nonlocal __file_size
            nonlocal l_c
            prev_time = time.time()
            try:
                if temp_file_size:
                    temp_output_file = open(resume, "a+b")
                else:
                    temp_output_file = tempfile.NamedTemporaryFile(
                        mode="w+b",
                        prefix="megapy_",
                        delete=False,
                        dir=default_tempfile_dir
                    )
                k_str = a32_to_str(k)
                counter = Counter.new(128, initial_value=((iv[0] << 32) + iv[1]) << 64)
                aes = AES.new(k_str, AES.MODE_CTR, counter=counter)
                mac_str = "\0" * 16
                mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str.encode("utf8"))
                iv_str = a32_to_str([iv[0], iv[1], iv[0], iv[1]])
                if temp_file_size:
                    if verbose:
                        print("\rresuming, fast forwarding aes counter", end="")
                    remaining = temp_file_size
                    _size = 65536
                    while remaining > 0:
                        print("\r", "resuming, fast forwarding, remaining", remaining, end="")
                        hm = min(_size, remaining)
                        aes.decrypt(b"\0"*hm)
                        remaining -= hm
                print("\rresumed, fast forward complete", end="")
                while index < l and not exploded:
                    try:
                        _file_size, chunk = result.pop(index)
                        if _file_size is Exception:
                            raise Exception("ran out of proxies, cannot download part {} bytes {}-{}".format(chunk[0], chunk[2], chunk[3]))
                        if isinstance(chunk, str):
                            _chunk = chunk
                            chunk = open(_chunk, "rb").read()
                            os.remove(_chunk)
                        l_c = len(chunk)
                        if verbose:
                            __file_size = _file_size
                            progress()
                            # print("got chunk", index+1, "out of", l, _file_size, l_c)
                        if _file_size != l_c:
                            open(os.path.join(tempfile.gettempdir(), "megapy", "csm-{}-{}-{}-{}.bin".format(file_id, index, _file_size, int(time.time()))), "wb").write(chunk)
                            raise Exception("chunk size mismatch")
                        completed += l_c
                        chunk = aes.decrypt(chunk)
                        temp_output_file.write(chunk)
                        if not temp_file_size:
                            encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
                            for i in range(0, l_c - 16, 16):
                                block = chunk[i:i + 16]
                                encryptor.encrypt(block)
                            if _file_size > 16:
                                i += 16
                            else:
                                i = 0
                            block = chunk[i:i + 16]
                            if len(block) % 16:
                                block += b"\0" * (16 - (len(block) % 16))
                            mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))
                        index += 1
                        open(resume_fp, "wb").write(temp_output_file.name.encode())
                        if verbose:
                            progress()
                            # print("completed chunks", index+1, "out of", l, completed/file_size*100)
                        prev_time = time.time()
                    except KeyError:
                        time.sleep(1/1000)
                        cur_time = time.time()
                        if cur_time-prev_time>1:
                            prev_time = cur_time
                            progress()
                            # print("handler listening chunk", index+1, "out of", l)
                    except Exception as e:
                        raise e
                if not exploded:
                    if not temp_file_size:
                        file_mac = str_to_a32(mac_str)
                        if (file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3]) != meta_mac:
                            raise ValueError("Mismatched mac")
                    output_path = Path(dest_path + file_path + file_name)
                    output_path.parent.mkdir(parents=True, exist_ok=True)
                    temp_output_file.close()
                    if verbose:
                        print("\rmoving", temp_output_file.name, "==>", output_path)
                    shutil.move(temp_output_file.name, output_path, copy_function=shutil.copyfile)
                    os.remove(resume_fp)
                    if verbose:
                        print("\rmoved", temp_output_file.name, "==>", output_path)
            except:
                print("\rhandler")
                traceback.print_exc()
                exploded = True
            done = True
        def job(i, j, a, b):
            nonlocal _i
            nonlocal exploded
            nonlocal downloaded
            if exploded:
                print("\r", "exploded", end="")
                return
            class _r:
                content = b""
                status_code = 404
                def close(self):
                    pass
            # open("log.log", "ab").write(("{}_{} start\n".format(proxies[j], i)).encode())
            temp_piece_fp = os.path.join(temp_pieces_fp, "{}_{}_{}.bin".format(i, a, b))
            try:
                proxy = {"all": proxies[j]} if proxies[j] else None
                if proxies[j] not in timeout_blacklist:
                    try:
                        if os.path.isfile(temp_piece_fp):
                            r = _r()
                            r.status_code = 200
                            r.content = open(temp_piece_fp, "rb").read()
                            os.remove(temp_piece_fp)
                        else:
                            if j not in cache:
                                if is_dir:
                                    cache[j] = self.get_file_data_in_folder(file_id, root_folder, proxy)["g"]+"/{}-{}"
                                else:
                                    cache[j] = self.get_file_data(file_id, proxy)["g"]+"/{}-{}"
                            url = cache[j].format(a, b)
                            r = requests.get(url, stream=True, timeout=5, proxies=proxy)
                    except:
                        traceback.print_exc()
                        r = _r()
                else:
                    print("\rproxy in blacklist")
                    r = _r()
                try:
                    c = r.content
                    if not c:
                        raise
                except:
                    status_code = r.status_code
                    r = _r()
                    if status_code != 200:
                        r.status_code = status_code
                if r.status_code != 200:
                    if stop_on_exception and abs((j+1)%len(proxies)-retry[i]) > 10:
                        r.close()
                        raise Exception("stop_on_exception", stop_on_exception, r.status_code)
                    if proxies[j] not in timeout_blacklist and r.status_code != 509:
                        timeout_blacklist.append(proxies[j])
                        # open("timeout_blacklist.json", "wb").write(json.dumps(timeout_blacklist).encode())
                    if (j+1)%len(proxies) != retry[i]:
                        if verbose:
                            print("\r", "failed to get chunk", i+1, "out of", l, a, b, ", trying proxy", j+1, "out of", len(proxies), end="")
                            print("\r", proxies[j], r.status_code, end="")
                        # open("log.log", "ab").write(("{}_{} end\n".format(proxies[j], i)).encode())
                        r.close()
                        # job(i, j+1, a, b)
                        job(i, (j+1)%len(proxies), a, b)
                        return
                    else:
                        result[i] = [Exception, [i,j,a,b]]
                        r.close()
                else:
                    if verbose:
                        _i = i
                        progress()
                        # print("getting chunk", i+1, "out of", l, a, b)
                    if i-index>parallel_threads*(2**3):
                        open(temp_piece_fp, "wb").write(c)
                        c = temp_piece_fp
                    result[i] = [b-a+1, c]
                    downloaded += b-a+1
                    if verbose:
                        _i = i
                        progress()
                        # print("downloaded chunks", i+1, "out of", l, a, b, downloaded/file_size*100)
                    r.close()
            except:
                print("\rjob")
                traceback.print_exc()
                exploded = True
            # open("log.log", "ab").write(("{}_{} end\n".format(proxies[j], i)).encode())
        p = threading.Thread(target=file_handler)
        p.daemon = True
        p.start()
        def get_proxy_load():
            return {j: len([0 for __ in _.threads if __.is_alive()]) for j, _ in enumerate(tws)}
        def get_min_load_proxy_j():
            min_j = []
            min_load = 999
            for _j, load in get_proxy_load().items():
                if load < min_load:
                    min_load = load
                    min_j.clear()
                if load <= min_load:
                    min_j.append(_j)
            if len(min_j) == 1 or min_load == 0:
                return min_j[0]
            else:
                return -1
        for i, (a, b) in enumerate(chunks):
            if a + b <= temp_file_size:
                index = i+1
                if verbose:
                    print("\rskipped chunk", i+1, a, a+b-1, end="")
                continue
            while True:
                # j = get_min_load_proxy_j()
                j = random.SystemRandom().choice(list(range(0, len(proxies))))
                if proxies[j] in timeout_blacklist:
                    j = -1
                if j == -1:
                    time.sleep(1/10)
                    print("\rretrying", i, j)
                else:
                    break
            # tws[j].add(job=def_template(job, i, j, a, a+b-1))
            # j = i%parallel_threads
            retry[i] = j
            tws[i%parallel_threads].add(job=def_template(job, i, j, a, a+b-1))
        [tw.wait() for tw in tws]
        while not done:
            time.sleep(1/10)
        if exploded:
            raise Exception("parallel download failed")
        else:
            os.rmdir(temp_pieces_fp)

    def get_file_name(self,
                       file_handle,
                       file_key=None):
        if file_key is None and file_handle.startswith("http"):
            file_handle, file_key = self._parse_url(file_handle).split("!")
        file_key = base64_to_a32(file_key)
        file_data = self._api_request({
            'a': 'g',
            'g': 1,
            'p': file_handle
        })
        k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
             file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
        if 'g' not in file_data:
            raise RequestError('File not accessible anymore')
        attribs = base64_url_decode(file_data['at'])
        attribs = decrypt_attr(attribs, k)
        file_name = attribs['n']
        return file_name

    def get_file_size(self,
                       file_handle,
                       file_key=None):
        if file_key is None and file_handle.startswith("http"):
            file_handle, file_key = self._parse_url(file_handle).split("!")
        file_key = base64_to_a32(file_key)
        file_data = self._api_request({
            'a': 'g',
            'g': 1,
            'p': file_handle
        })
        k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
             file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
        iv = file_key[4:6] + (0, 0)
        meta_mac = file_key[6:8]
        if 'g' not in file_data:
            raise RequestError('File not accessible anymore')
        return file_data['s']

    def get_file_link(self,
                       file_handle,
                       file_key=None):
        if file_key is None and file_handle.startswith("http"):
            file_handle, file_key = self._parse_url(file_handle).split("!")
        file_key = base64_to_a32(file_key)
        file_data = self._api_request({
            'a': 'g',
            'g': 1,
            'p': file_handle
        })
        k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
             file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
        iv = file_key[4:6] + (0, 0)
        meta_mac = file_key[6:8]
        if 'g' not in file_data:
            raise RequestError('File not accessible anymore')
        return file_data['g']

    def _download_file(self,
                       file_handle,
                       file_key,
                       dest_path=None,
                       dest_filename=None,
                       is_public=False,
                       file=None,
                       serial_download=True,
                       parallel_download=False,
                       parallel_threads=2**4,
                       verbose=True,
                       proxies=None,
                       stop_on_exception=False,
                       url=None):
        if file is None:
            if is_public:
                file_key = base64_to_a32(file_key)
                file_data = self._api_request({
                    "a": "g",
                    "g": 1,
                    "p": file_handle
                })
            else:
                file_data = self._api_request({
                    "a": "g",
                    "g": 1,
                    "n": file_handle
                })

            k = (file_key[0] ^ file_key[4], file_key[1] ^ file_key[5],
                 file_key[2] ^ file_key[6], file_key[3] ^ file_key[7])
            iv = file_key[4:6] + (0, 0)
            meta_mac = file_key[6:8]
        else:
            file_data = self._api_request({"a": "g", "g": 1, "n": file["h"]})
            k = file["k"]
            iv = file["iv"]
            meta_mac = file["meta_mac"]
        if parallel_download:
            return self.parallel_download(file_handle, url, dest_path, dest_filename, file_data, k, iv, meta_mac, parallel_threads, verbose, proxies, stop_on_exception)
        elif serial_download:
            if "g" not in file_data:
                raise RequestError("File not accessible anymore")
            return self.download_file(dest_path, dest_filename, file_data, k, iv, meta_mac, verbose, proxies)
        raise Exception("please specify a download method: serial_download or parallel_download")

    def serial_download(self, file_url, k, iv, meta_mac, file_size, dest_path, file_path, file_name, proxies):
        proxy = random.SystemRandom().choice(proxies)
        proxy = random.SystemRandom().choice(proxies)
        r = requests.get(file_url, stream=True, proxies={"all": proxy} if proxy else None)
        if r.status_code == 509:
            raise Exception("transfer quota exceeded. please try again with other IP", proxy)
        input_file = r.raw
        temp_output_file = tempfile.NamedTemporaryFile(
            mode="w+b",
            prefix="megapy_",
            delete=False
        )
        k_str = a32_to_str(k)
        counter = Counter.new(128, initial_value=((iv[0] << 32) + iv[1]) << 64)
        aes = AES.new(k_str, AES.MODE_CTR, counter=counter)
        mac_str = "\0" * 16
        mac_encryptor = AES.new(k_str, AES.MODE_CBC, mac_str.encode("utf8"))
        iv_str = a32_to_str([iv[0], iv[1], iv[0], iv[1]])
        for chunk_start, chunk_size in get_chunks(file_size):
            chunk = input_file.read(chunk_size)
            chunk = aes.decrypt(chunk)
            temp_output_file.write(chunk)
            encryptor = AES.new(k_str, AES.MODE_CBC, iv_str)
            for i in range(0, len(chunk) - 16, 16):
                block = chunk[i:i + 16]
                encryptor.encrypt(block)
            if file_size > 16:
                i += 16
            else:
                i = 0
            block = chunk[i:i + 16]
            if len(block) % 16:
                block += b"\0" * (16 - (len(block) % 16))
            mac_str = mac_encryptor.encrypt(encryptor.encrypt(block))
            file_info = os.stat(temp_output_file.name)
            logger.info("%s of %s downloaded", file_info.st_size, file_size)
        file_mac = str_to_a32(mac_str)
        if (file_mac[0] ^ file_mac[1], file_mac[2] ^ file_mac[3]) != meta_mac:
            raise ValueError("Mismatched mac")
        output_path = Path(dest_path + file_path + file_name)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        temp_output_file.close()
        shutil.move(temp_output_file.name, output_path, copy_function=shutil.copyfile)
        return output_path

    def download_file(self, dest_path: str, dest_filename: str, file_data: dict, k: tuple, iv: tuple, meta_mac: tuple, verbose=True, proxies=None) -> str:
        proxies = proxies or [None]
        file_url = file_data["g"]
        file_size = file_data["s"]
        file_path = file_data["path"][1:] if "path" in file_data else ""
        attribs = base64_url_decode(file_data["at"])
        attribs = decrypt_attr(attribs, k)
        if dest_filename is not None:
            file_name = dest_filename
        else:
            file_name = attribs["n"]
        if verbose:
            print("\r", file_name, file_size, flush=True)
        if not dest_path:
            dest_path = ""
        else:
            dest_path += "/"
        return self.serial_download(file_url, k, iv, meta_mac, file_size, dest_path, file_path, file_name, proxies)

    def download_url(self, url, dest_path=None, dest_filename=None, serial_download=True, parallel_download=False, parallel_threads=2**4, verbose=True, proxies=None, stop_on_exception=False):
        """
        Download a file by it's public url
        """
        proxies = proxies or [None]
        path = self._parse_url(url).split('!')
        file_id = path[0]
        file_key = path[1]
        return self._download_file(
            url=url,
            file_handle=file_id,
            file_key=file_key,
            dest_path=dest_path,
            dest_filename=dest_filename,
            is_public=True,
            serial_download=serial_download,
            parallel_download=parallel_download,
            parallel_threads=parallel_threads,
            verbose=verbose,
            proxies=proxies,
            stop_on_exception=stop_on_exception
        )

